from typing import Optional, Dict, Any, Union

import requests

from simple_smartsheet import constants
from simple_smartsheet import exceptions
from simple_smartsheet.models.sheet import SheetsCRUD
from simple_smartsheet.models.extra import Result
from simple_smartsheet.types import JSONType


class Smartsheet:
    """Smartsheet API class that provides a way to interact with Smartsheet objects.

    Attributes:
        token: Smartsheet API token, obtained in Personal Settings -> API access
        session: requests.Session object which stores headers based on the token
        sheets: SheetsCRUD object which provides methods to interact with Sheets
    """

    API_HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}

    def __init__(self, token: str) -> None:
        self.session = requests.Session()
        self.token = token
        self.sheets = SheetsCRUD(self)

    @property
    def token(self) -> str:
        return self._token

    @token.setter
    def token(self, value) -> None:
        self._token = value
        # when the token is changed, update headers too
        self._update_headers()

    def _update_headers(self) -> None:
        """Updates HTTP Headers with the token"""
        headers = {"Authorization": f"Bearer {self.token}", **self.API_HEADERS}
        self.session.headers.update(headers)

    @staticmethod
    def get_endpoint_url(endpoint: str) -> str:
        """Build a full API url based on the relative API path.

        For example:
            get_endpoint_url('/sheets') -> 'https://api.smartsheet.com/2.0/sheets'
        """
        return constants.API_ROOT + endpoint

    def get(
        self,
        endpoint: str,
        path: Optional[str] = "data",
        params: Optional[Dict[str, Any]] = None,
    ) -> JSONType:
        """Performs HTTP GET on the endpoint

        Args:
            endpoint: relative API endpoint, for example '/sheets'
            path: in response extract the data under the specific path.
                Default - 'data'. Specify None if not needed
            params: HTTP query params dictionary

        Returns:
            JSON data from the response, under the specific key.
        """
        url = self.get_endpoint_url(endpoint)
        response = self.session.get(url, params=params)
        if response.ok:
            if response.text:
                response_json = response.json()
                if path:
                    if path in response_json:
                        return response_json[path]
                    raise AttributeError(
                        f"Response from GET {url} does not contain key {path!r}"
                    )
                else:
                    return response_json
            else:
                return {}
        else:
            raise exceptions.SmartsheetHTTPError.from_response(response)

    def post(
        self, endpoint: str, data: Optional[JSONType] = None, result_obj: bool = True
    ) -> Union[Result, Dict[str, Any], None]:
        """Performs HTTP POST on the endpoint

        Args:
            endpoint: relative API endpoint, for example '/sheets'
            data: dictionary or list with data that is going to be sent as JSON
            result_obj: whether to convert received JSON response to Result object

        Returns:
            Result object
        """
        url = self.get_endpoint_url(endpoint)

        if data:
            response = self.session.post(url, json=data)
        else:
            response = self.session.post(url)
        if response.ok:
            if response.text:
                json_response = response.json()
                if result_obj:
                    result = Result.load(json_response)
                    return result
                else:
                    return json_response
            else:
                return None
        else:
            raise exceptions.SmartsheetHTTPError.from_response(response)

    def put(self, endpoint: str, data: JSONType) -> Optional[Result]:
        """Performs HTTP PUT on the endpoint

        Args:
            endpoint: relative API endpoint, for example '/sheets'
            data: dictionary or list with data that is going to be sent as JSON

        Returns:
            Result object
        """
        url = self.get_endpoint_url(endpoint)

        response = self.session.put(url, json=data)
        if response.ok:
            result = Result.load(response.json())
            return result
        else:
            raise exceptions.SmartsheetHTTPError.from_response(response)

    def delete(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Result:
        """Performs HTTP DELETE on the endpoint

        Args:
            endpoint: relative API endpoint, for example '/sheets'
            params: HTTP query params dictionary

        Returns:
            Result object
        """
        url = self.get_endpoint_url(endpoint)

        response = self.session.delete(url, params=params)
        if response.ok:
            result = Result.load(response.json())
            return result
        else:
            raise exceptions.SmartsheetHTTPError.from_response(response)



import json
from typing import Optional

from requests import Response

from simple_smartsheet import models


class SmartsheetError(Exception):
    pass


class SmartsheetHTTPError(SmartsheetError):
    def __init__(
        self,
        http_response_code: int,
        error: "models.extra.Error" = None,
        message: str = "",
    ):
        result_msg_items = [f"HTTP response code {http_response_code}"]
        if error is not None:
            result_msg_items.append(f"Error code {error.error_code}")
            if error.message:
                message = error.message

        if message:
            result_msg_items.append(message)

        result_msg = " - ".join(result_msg_items)
        super().__init__(result_msg)

        self.http_response_code = http_response_code
        self.message = message
        self.error = error

    @classmethod
    def from_response(cls, response: Response):
        http_response_code = response.status_code
        error: Optional["models.extra.Error"] = None
        message = ""
        if response.text:
            try:
                error = models.extra.Error.load(response.json())
            except json.JSONDecodeError:
                message = response.text

        if 400 <= http_response_code < 500:
            return SmartsheetHTTPClientError(http_response_code, error, message)
        elif 500 <= http_response_code < 600:
            return SmartsheetHTTPServerError(http_response_code, error, message)
        else:
            return cls(http_response_code, error, message)


class SmartsheetHTTPClientError(SmartsheetHTTPError):
    pass


class SmartsheetHTTPServerError(SmartsheetHTTPError):
    pass


class SmartsheetObjectNotFound(SmartsheetError):
    pass



import logging
from collections import defaultdict
from datetime import datetime
from typing import (
    Optional,
    Dict,
    List,
    ClassVar,
    Type,
    Sequence,
    Tuple,
    Any,
    Union,
    cast,
)

import attr
from marshmallow import fields


from simple_smartsheet.models.base import Schema, CoreSchema, Object, CoreObject, CRUD
from simple_smartsheet.models.column import Column, ColumnSchema
from simple_smartsheet.models.row import Row, RowSchema
from simple_smartsheet.models.cell import Cell
from simple_smartsheet.models.extra import Result
from simple_smartsheet.types import IndexKeysDict, IndexesType


logger = logging.getLogger(__name__)


class UserSettingsSchema(Schema):
    critical_path_enabled = fields.Bool(data_key="criticalPathEnabled")
    display_summary_tasks = fields.Bool(data_key="displaySummaryTasks")


@attr.s(auto_attribs=True, repr=False, kw_only=True)
class UserSettings(Object):
    critical_path_enabled: bool
    display_summary_tasks: bool


class WorkspaceSchema(Schema):
    id = fields.Int()
    name = fields.Str()


@attr.s(auto_attribs=True, repr=False, kw_only=True)
class Workspace(Object):
    id: int
    name: str


class SheetSchema(CoreSchema):
    """Marshmallow Schema for Smartsheet Sheet object

    Additional details about fields can be found here:
    http://smartsheet-platform.github.io/api-docs/#sheets

    """

    id = fields.Int()
    name = fields.Str()
    access_level = fields.Str(data_key="accessLevel")
    permalink = fields.Str()
    favorite = fields.Bool()
    created_at = fields.DateTime(data_key="createdAt")
    modified_at = fields.DateTime(data_key="modifiedAt")

    version = fields.Int()
    total_row_count = fields.Int(data_key="totalRowCount")
    effective_attachment_options = fields.List(
        fields.Str(), data_key="effectiveAttachmentOptions"
    )
    gantt_enabled = fields.Bool(data_key="ganttEnabled")
    dependencies_enabled = fields.Bool(data_key="dependenciesEnabled")
    resource_management_enabled = fields.Bool(data_key="resourceManagementEnabled")
    cell_image_upload_enabled = fields.Bool(data_key="cellImageUploadEnabled")
    user_settings = fields.Nested(UserSettingsSchema, data_key="userSettings")

    columns = fields.Nested(ColumnSchema, many=True)
    rows = fields.Nested(RowSchema, many=True)
    workspace = fields.Nested(WorkspaceSchema)


@attr.s(auto_attribs=True, repr=False, kw_only=True)
class Sheet(CoreObject):
    """Represent Smartsheet Sheet object

    Additional details about fields can be found here:
    http://smartsheet-platform.github.io/api-docs/#sheets

    Extra attributes:
        row_rum_to_row: mapping of row number to Row object
        row_id_to_row: mapping of row id to Row object
        column_title_to_column: mapping of column title to Column object
        column_id_to_column: mapping of column id to Column object
        schema: reference to SheetSchema
    """

    name: str
    id: Optional[int] = None
    access_level: Optional[str] = None
    permalink: Optional[str] = None
    favorite: Optional[bool] = None
    created_at: Optional[datetime] = None
    modified_at: Optional[datetime] = None

    version: Optional[int] = None
    total_row_count: Optional[int] = None
    effective_attachment_options: List[str] = attr.Factory(list)
    gantt_enabled: Optional[bool] = None
    dependencies_enabled: Optional[bool] = None
    resource_management_enabled: Optional[bool] = None
    cell_image_upload_enabled: Optional[bool] = None
    user_settings: Optional[UserSettings] = None

    columns: List[Column] = attr.Factory(list)
    rows: List[Row] = attr.Factory(list)
    workspace: Optional[Workspace] = None

    row_num_to_row: Dict[int, Row] = attr.Factory(dict)
    row_id_to_row: Dict[int, Row] = attr.Factory(dict)
    column_title_to_column: Dict[str, Column] = attr.Factory(dict)
    column_id_to_column: Dict[int, Column] = attr.Factory(dict)

    index_keys: List[IndexKeysDict] = attr.Factory(list)
    index_key_to_unique: Dict[Tuple[str, ...], bool] = attr.Factory(dict)
    indexes: IndexesType = attr.Factory(lambda: defaultdict(dict))

    schema: ClassVar[Type[SheetSchema]] = SheetSchema

    def __attrs_post_init__(self) -> None:
        self.update_index(deserealize_cell_values=True)

    def update_index(self, deserealize_cell_values: bool = False) -> None:
        """Updates columns and row indices for quick lookup"""
        for index_key_dict in self.index_keys:
            columns = index_key_dict["columns"]
            unique = index_key_dict["unique"]
            self.index_key_to_unique[tuple(sorted(columns))] = unique

        self.update_column_index()
        self.update_row_index(self.index_key_to_unique, deserealize_cell_values)

    def update_column_index(self) -> None:
        """Updates columns index for quick lookup by title and ID"""
        self.column_title_to_column.clear()
        self.column_id_to_column.clear()

        for column in self.columns:
            if column.id is None:
                continue
            self.column_id_to_column[column.id] = column

            column_title = column.title
            if column_title is None:
                continue
            if column_title in self.column_title_to_column:
                logger.info(
                    "Column with the title %s is already present in the index",
                    column_title,
                )
            self.column_title_to_column[column_title] = column

    def update_row_index(
        self,
        index_keys: Dict[Tuple[str, ...], bool],
        deserealize_cell_values: bool = False,
    ) -> None:
        """Updates row index for quick lookup by row number and ID"""
        self.row_num_to_row.clear()
        self.row_id_to_row.clear()

        for row in self.rows:
            self.row_num_to_row[row.num] = row
            self.row_id_to_row[row.id] = row
            row.update_index(self, index_keys, deserealize_cell_values)

    def get_row(
        self,
        row_num: Optional[int] = None,
        row_id: Optional[int] = None,
        filter: Optional[Dict[str, Any]] = None,
    ) -> Optional[Row]:
        """Returns Row object by row number or ID

        Either row_num or row_id must be provided

        Args:
            row_num: row number
            row_id: row id
            filter: a dictionary with column title to value
        mappings in the same order as index was built. Index must be unique.

        Returns:
            Row object
        """
        if row_num is not None:
            return self.row_num_to_row.get(row_num)
        elif row_id is not None:
            return self.row_id_to_row.get(row_id)
        elif filter is not None:
            index_key, query = zip(*sorted(filter.items()))
            unique = self.index_key_to_unique.get(index_key)
            if unique is None:
                raise ValueError("Index %s was not built", index_key)
            elif not unique:
                raise ValueError(
                    "Index %s is non-unique and lookup will potentially "
                    "return multiple rows, use get_rows method instead",
                    index_key,
                )
            index = cast(Dict[Tuple, Row], self.indexes[index_key])
            return index[query]
        else:
            raise ValueError("Either row_num or row_id argument should be provided")

    def get_rows(self, filter: Dict[str, Any]) -> [List[Row]]:
        """Returns Row objects by index query

        Args:
            filter: a dictionary or ordered dictionary with column title to value
        mappings in the same order as index was built. Index must be non-unique.

        Returns:
            Row object
        """
        index_key, query = zip(*sorted(filter.items()))
        unique = self.index_key_to_unique.get(index_key)
        if unique is None:
            raise ValueError("Index %s was not built", index_key)
        elif unique:
            index = cast(Dict[Tuple, Row], self.indexes[index_key])
            result = index.get(query)
            if result is not None:
                return [result]
            else:
                return []
        else:
            index = cast(Dict[Tuple, List[Row]], self.indexes[index_key])
            return index.get(query, [])

    def get_column(
        self, column_title: Optional[str] = None, column_id: Optional[int] = None
    ) -> Optional[Column]:
        """Returns Column object by column title or ID

        Either column_title or column_id must be provided

        Args:
            column_title: column title (case-sensitive)
            column_id: column id

        Returns:
            Column object
        """
        if column_title is not None:
            return self.column_title_to_column.get(column_title)
        elif column_id is not None:
            return self.column_id_to_column.get(column_id)
        else:
            raise ValueError(
                "Either column_title or column_id argument should be provided"
            )

    def add_rows(self, rows: Sequence[Row]) -> Result:
        """Adds several rows to the smartsheet.

        Sheet must have api attribute set. It is automatically set when method
            Smartsheet.sheets.get() is used
        Every row must have either location-specifier attributes or row number set
        More details: http://smartsheet-platform.github.io/api-docs/#add-rows

        Args:
            rows: sequence of Row objects

        Returns:
            Result object
        """
        if self.api is None:
            raise ValueError("To use this method, api attribute must be set")
        include_fields = (
            "parent_id",
            "sibling_id",
            "above",
            "indent",
            "outdent",
            "to_bottom",
            "to_top",
            "expanded",
            "format",
            "cells.column_id",
            "cells.formula",
            "cells.value",
            "cells.hyperlink",
            "cells.link_in_from_cell",
            "cells.strict",
            "cells.format",
            "cells.image",
            "cells.override_validation",
            "locked",
        )
        data = []
        schema = RowSchema(only=include_fields)
        for row in rows:
            new_row = row.copy(deep=False)
            new_row.cells = [
                cell
                for cell in row.cells
                if cell.value is not None or cell.formula is not None
            ]
            data.append(schema.dump(new_row))
        return self.api.post(f"/sheets/{self.id}/rows", data=data)

    def add_row(self, row: Row) -> Result:
        """Adds a single row to the smartsheet.

        Sheet must have api attribute set. It is automatically set when method
            Smartsheet.sheets.get() is used
        A row must have either location-specifier attributes or row number set
        More details: http://smartsheet-platform.github.io/api-docs/#add-rows

        Args:
            row: Row object

        Returns:
            Result object
        """
        return self.add_rows([row])

    def update_rows(self, rows: Sequence[Row]) -> Result:
        """Updates several rows in the Sheet.

        Sheet must have api attribute set. It is automatically set when method
            Smartsheet.sheets.get() is used
        More details: http://smartsheet-platform.github.io/api-docs/#update-rows

        Args:
            rows: sequence of Row objects

        Returns:
            Result object
        """
        if self.api is None:
            raise ValueError("To use this method, api attribute must be set")
        include_fields = (
            "id",
            "parent_id",
            "above",
            "indent",
            "outdent",
            "to_bottom",
            "to_top",
            "expanded",
            "format",
            "cells.column_id",
            "cells.formula",
            "cells.value",
            "cells.hyperlink",
            "cells.link_in_from_cell",
            "cells.strict",
            "cells.format",
            "cells.image",
            "cells.override_validation",
            "locked",
        )
        data = []
        schema = RowSchema(only=include_fields)
        for row in rows:
            new_row = row.copy(deep=False)
            new_row.cells = [
                cell
                for cell in row.cells
                if cell.value is not None or cell.formula is not None
            ]
            data.append(schema.dump(new_row))
        return self.api.put(f"/sheets/{self.id}/rows", data=data)

    def update_row(self, row: Row) -> Result:
        """Updates a single row in the Sheet.

        Sheet must have api attribute set. It is automatically set when method
            Smartsheet.sheets.get() is used
        More details: http://smartsheet-platform.github.io/api-docs/#update-rows

        Args:
            row: Row object

        Returns:
            Result object
        """
        return self.update_rows([row])

    def delete_rows(self, row_ids: Sequence[int]) -> Result:
        """Deletes several rows in the Sheet.

        Rows are identified by ids.

        Args:
            row_ids: sequence of row ids

        Returns:
            Result object
        """
        if self.api is None:
            raise ValueError("To use this method, api attribute must be set")
        endpoint = f"/sheets/{self.id}/rows"
        params = {"ids": ",".join(str(row_id) for row_id in row_ids)}
        return self.api.delete(endpoint, params=params)

    def delete_row(self, row_id: int) -> Result:
        """Deletes a single row in the Sheet specified by ID.

        Args:
            row_id: Row id

        Returns:
            Result object
        """
        return self.delete_rows([row_id])

    def sort_rows(self, order: List[Dict[str, Any]]) -> "Sheet":
        """Sorts rows in the sheet with the specified order


        Args:
            order: List of dictionaries containing column_title or column_id and
                (optional) descending bool (default is ascending). Example:
                [
                    {"column_title": "Birth date", "descending": True},
                    {"column_title": "Full Name"}
                ]

        Returns:
            Sheet object
        """
        # TODO: add validation schema for sorting order
        normalized_order = []
        for item in order:
            normalized_item = {}
            if "column_id" in item:
                normalized_item["columnId"] = item["column_id"]
            elif "column_title" in item:
                column_title = item["column_title"]
                column = self.get_column(column_title)
                normalized_item["columnId"] = column.id
            else:
                raise ValueError(
                    "Sorting key must have either column_id or column_title"
                )

            descending = item.get("descending", False)
            if descending:
                normalized_item["direction"] = "DESCENDING"
            else:
                normalized_item["direction"] = "ASCENDING"
            normalized_order.append(normalized_item)

        data = {"sortCriteria": normalized_order}
        endpoint = f"/sheets/{self.id}/sort"
        response = self.api.post(endpoint, data, result_obj=False)
        sheet = self.load(response)
        sheet.api = self.api
        return sheet

    def make_cell(
        self, column_title: str, field_value: Union[float, str, datetime, None]
    ) -> Cell:
        """Creates a Cell object for an existing column

        Args:
            column_title: title of an existing column
            field_value: value of the cell

        Returns:
            Cell object
        """
        column = self.get_column(column_title)
        if column is None:
            raise ValueError(
                "A column with the title %s does not exist in this sheet", column_title
            )
        cell = Cell(column_id=column.id, value=field_value)
        return cell

    def make_cells(
        self, fields: Dict[str, Union[float, str, datetime, None]]
    ) -> List[Cell]:
        """Create a list of Cell objects from dictionary

        Args:
            fields: dictionary where key is a column title and value is a cell value

        Returns:
            list of Cell objects
        """
        result: List[Cell] = []
        for column_title, field_value in fields.items():
            result.append(self.make_cell(column_title, field_value))
        return result

    def as_list(self) -> List[Dict[str, Union[float, str, datetime, None]]]:
        """Returns a list of dictionaries with column titles and cell values"""
        return [row.as_dict() for row in self.rows]


class SheetsCRUD(CRUD[Sheet]):
    base_url = "/sheets"
    factory = Sheet
    default_path = "sheets"

    create_include_fields = (
        "name",
        "columns.primary",
        "columns.title",
        "columns.type",
        "columns.auto_number_format",
        "columns.options",
        "columns.symbol",
        "columns.system_column_type",
        "columns.width",
    )



import os
from typing import Optional, Sequence, Any

from marshmallow import EXCLUDE, RAISE


def get_unknown_field_handling(debug: bool) -> str:
    if debug:
        return RAISE
    else:
        return EXCLUDE


def is_env_var(env_var: str) -> bool:
    env_var_str = os.getenv(env_var, "").lower()
    return env_var_str in ("yes", "true", "y", "1")


def is_debug() -> bool:
    return is_env_var("SIMPLE_SMARTSHEET_DEBUG")


def create_repr(obj: Any, attrs: Optional[Sequence[str]] = None):
    if attrs is None:
        attrs = obj.__dict__.keys()
    attrs_repr = ", ".join(f"{attr}={getattr(obj, attr)!r}" for attr in attrs)
    return f"{obj.__class__.__qualname__}({attrs_repr})"



from typing import Union, Dict, List, Any, Tuple, DefaultDict, TYPE_CHECKING

from mypy_extensions import TypedDict

if TYPE_CHECKING:
    from simple_smartsheet.models import Row  # noqa

JSONType = Union[Dict[str, Any], List[Dict[str, Any]]]

IndexKeysDict = TypedDict("IndexKeysDict", {"columns": Tuple[str, ...], "unique": bool})
IndexesKeysType = List[IndexKeysDict]
IndexesType = DefaultDict[
    Tuple[str, ...], Dict[Tuple[Any, ...], Union["Row", List["Row"]]]
]



import logging
from datetime import datetime
from typing import Optional, List, TYPE_CHECKING, Dict, Tuple, Any, ClassVar, Type, cast

import attr
from marshmallow import fields

from simple_smartsheet import utils
from simple_smartsheet.models.base import Schema, Object
from simple_smartsheet.models.cell import Cell, CellSchema
from simple_smartsheet.models.column import Column, ColumnSchema

if TYPE_CHECKING:
    from simple_smartsheet.models.sheet import Sheet

logger = logging.getLogger(__name__)


class RowSchema(Schema):
    id = fields.Int()
    sheet_id = fields.Int(data_key="sheetId")
    access_level = fields.Str(data_key="accessLevel")
    attachments = fields.List(fields.Field())  # TODO: Attachment object
    cells = fields.Nested(CellSchema, many=True)
    columns = fields.Nested(ColumnSchema, many=True)
    conditional_format = fields.Str(data_key="conditionalFormat")
    created_at = fields.DateTime(data_key="createdAt")
    created_by = fields.Field(data_key="createdBy")  # TODO: User object
    discussions = fields.List(fields.Field())  # TODO: Discussion object
    expanded = fields.Bool()
    filtered_out = fields.Bool(data_key="filteredOut")
    format = fields.Str()
    in_critical_path = fields.Bool(data_key="inCriticalPath")
    locked = fields.Bool()
    locked_for_user = fields.Bool(data_key="lockedForUser")
    modified_at = fields.DateTime(data_key="modifiedAt")
    modified_by = fields.Field(data_key="modifiedBy")  # TODO: User object
    num = fields.Int(data_key="rowNumber")
    permalink = fields.Str()
    version = fields.Int()

    # location-specifier attributes
    parent_id = fields.Int(data_key="parentId")
    sibling_id = fields.Int(data_key="siblingId")
    above = fields.Bool()
    indent = fields.Int()
    outdent = fields.Int()
    to_bottom = fields.Bool(data_key="toBottom")
    to_top = fields.Bool(data_key="toTop")


@attr.s(auto_attribs=True, repr=False, kw_only=True)
class Row(Object):
    id: Optional[int] = None
    sheet_id: Optional[int] = None
    access_level: Optional[str] = None
    attachments: List[Any] = attr.Factory(list)
    cells: List[Cell] = attr.Factory(list)
    columns: List[Column] = attr.Factory(list)
    conditional_format: Optional[str] = None
    created_at: Optional[datetime] = None
    created_by: Optional[Any] = None
    discussions: List[Any] = attr.Factory(list)
    expanded: Optional[bool] = None
    filtered_out: Optional[bool] = None
    format: Optional[str] = None
    in_critical_path: Optional[bool] = None
    locked: Optional[bool] = None
    locked_for_user: Optional[bool] = None
    modified_at: Optional[datetime] = None
    modified_by: Optional[Any] = None
    num: Optional[int] = None
    permalink: Optional[str] = None
    version: Optional[int] = None

    # location-specified attributes
    parent_id: Optional[int] = None
    sibling_id: Optional[int] = None
    above: Optional[bool] = None
    indent: Optional[int] = None
    outdent: Optional[int] = None
    to_bottom: Optional[bool] = None
    to_top: Optional[bool] = None

    # index
    column_title_to_cell: Dict[str, Cell] = attr.Factory(dict)
    column_id_to_cell: Dict[int, Cell] = attr.Factory(dict)

    schema: ClassVar[Type[RowSchema]] = RowSchema

    def __repr__(self) -> str:
        return utils.create_repr(self, ["id", "num"])

    def update_index(
        self,
        sheet: "Sheet",
        index_key_to_unique: Dict[Tuple[str, ...], bool],
        deserealize_cell_values: bool = False,
    ) -> None:
        self.column_title_to_cell.clear()
        self.column_id_to_cell.clear()

        for cell in self.cells:
            column_id = cell.column_id
            if column_id is None:
                continue
            column = sheet.get_column(column_id=column_id)
            column_title = column.title

            self.column_id_to_cell[column_id] = cell
            if column_title is not None:
                self.column_title_to_cell[column_title] = cell

            if deserealize_cell_values:
                cell.deserealize_value(self, column)

        for index_key, unique in index_key_to_unique.items():
            index = sheet.indexes[index_key]
            key = tuple(self.get_cell(column_title).value for column_title in index_key)
            if unique:
                index[key] = self
            else:
                container = cast(List["Row"], index.setdefault(key, []))
                container.append(self)

    def get_cell(
        self, column_title: Optional[str] = None, column_id: Optional[int] = None
    ) -> Cell:
        if column_title is not None:
            return self.column_title_to_cell[column_title]
        elif column_id is not None:
            return self.column_id_to_cell[column_id]
        else:
            raise ValueError(
                "Either column_title or column_id argument should be provided"
            )

    def as_dict(self) -> Dict[str, Any]:
        """Returns a dictionary of column title to cell value"""
        return {
            column_title: cell.value
            for column_title, cell in self.column_title_to_cell.items()
        }



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['simple_smartsheet', 'simple_smartsheet.models']

package_data = \
{'': ['*']}

install_requires = \
['attrs',
 'cattrs',
 'marshmallow>=3.0.0rc1,<4.0.0',
 'mypy-extensions',
 'requests']

setup_kwargs = {
    'name': 'simple-smartsheet',
    'version': '0.1.11',
    'description': 'Python library to interact with Smartsheet API',
    'long_description': '## Simple Smartsheet\nPython library to interact with Smartsheet API easily\n\n### Installation\nRequires Python 3.6+  \n`pip install simple-smartsheet`\n\n### Why not smartsheet-python-sdk\n`smartsheet-python-sdk` has very wide object coverage and maps to Smartsheet API very nicely, but it does not have any additional features (for example, easy access to cells by column titles).  \n`simple-smartsheet` library is focused on user experience first in expense of feature coverage. \nAs of now, you can only interact with Sheets and nested objects (rows, columns, cells).\n\n### Usage\n```python\nfrom datetime import date\nfrom pprint import pprint\n\nfrom simple_smartsheet import Smartsheet\nfrom simple_smartsheet.models import Sheet, Column, Row, Cell\n\nTOKEN = "my-secret-token"\nsmartsheet = Smartsheet(TOKEN)\n\n# creating new Sheet\nnew_sheet = Sheet(\n    name="My New Sheet",\n    columns=[\n        Column(primary=True, title="Full Name", type="TEXT_NUMBER"),\n        Column(title="Number of read books", type="TEXT_NUMBER"),\n        Column(title="Birth date", type="DATE"),\n    ],\n)\n\n# print the sheet object as a dictionary which will be used in REST API\npprint(new_sheet.dump())\n\n# adding the sheet via API\nsmartsheet.sheets.create(new_sheet)\n\n# getting a simplified view of sheets\nsheets = smartsheet.sheets.list()\npprint(sheets)\n\n# getting the sheet by name\nsheet = smartsheet.sheets.get("My New Sheet")\n\n# printing the sheet object attributes\npprint(sheet.__dict__)\n# or printing the sheet object as a dictionary which will be used in REST API\npprint(sheet.dump())\n\n# getting columns details by column title (case-sensitive)\nfull_name_column = sheet.get_column("Full Name")\npprint(full_name_column.__dict__)\nnum_books_column = sheet.get_column("Number of read books")\npprint(num_books_column.__dict__)\n\n# adding rows (cells created using different ways):\nsheet.add_rows(\n    [\n        Row(\n            to_top=True,\n            cells=[\n                Cell(column_id=full_name_column.id, value="Alice Smith"),\n                Cell(column_id=num_books_column.id, value=5),\n            ],\n        ),\n        Row(\n            to_top=True,\n            cells=sheet.make_cells(\n                {"Full Name": "Bob Lee", "Number of read books": 2}\n            )\n        ),\n        Row(\n            to_top=True,\n            cells=[\n                sheet.make_cell("Full Name", "Charlie Brown"),\n                sheet.make_cell("Number of read books", 1),\n                sheet.make_cell("Birth date", date(1990, 1, 1)),\n            ],\n        ),\n    ]\n)\n\n# sort rows now by column "Full Name" descending / returns updated sheet\nsheet = sheet.sort_rows([{"column_title": "Full Name", "descending": True}])\n\n# or getting an updated sheet again\n# sheet = smartsheet.sheets.get("My New Sheet")\nprint("\\nSheet after adding rows:")\n# all sheet attributes\npprint(sheet.__dict__)\n# or just a list of dictionaries containing column titles and values\npprint(sheet.as_list())\n\n# getting a specific cell and updating it:\nrow_id_to_delete = None\nrows_to_update = []\nfor row in sheet.rows:\n    full_name = row.get_cell("Full Name").value\n    num_books = row.get_cell("Number of read books").value\n    print(f"{full_name} has read {num_books} books")\n    if full_name.startswith("Charlie"):\n        num_books_cell = row.get_cell("Number of read books")\n        num_books_cell.value += 1\n        rows_to_update.append(row)\n    elif full_name.startswith("Bob"):\n        row_id_to_delete = row.id  # used later\n\n# update rows\nsheet.update_rows(rows_to_update)\n# or a single row\n# sheet.update_rows(rows_to_update[0])\n\n# getting an updated sheet\nsheet = smartsheet.sheets.get("My New Sheet")\nprint("\\nSheet after updating rows:")\npprint(sheet.as_list())\n\n# deleting row by id\nsheet.delete_row(row_id_to_delete)\n\n# getting an updated sheet\nsheet = smartsheet.sheets.get("My New Sheet")\nprint("\\nSheet after deleting rows:")\npprint(sheet.as_list())\n\n# deleting Sheet\n# sheet = smartsheet.sheets.delete(\'My New Sheet\')\nsheets = smartsheet.sheets.list()\npprint(sheets)\n```\n\n### Docs\nWhile a separate docs page is work in progress, available public API is described here\n#### Class `simple_smartsheet.Smartsheet`\nThis class a main entry point for the library  \nMethods:\n  * `def __init__(token: str)`: constructor for the class\n  \nAttributes:\n  * `token`: Smartsheet API token, obtained in Personal Settings -> API access\n  * `session`: requests.Session object which stores headers based on the token\n  * `sheets`: `simple_smartsheet.models.sheet.SheetsCRUD` object which provides methods to interact with Sheets\n  \n#### Class `simple_smartsheet.models.sheet.SheetsCRUD`\nMethods:\n  * `def get(name: Optional[str], id: Optional[int], index_keys: Optional[Dict[str, Any]])`: fetches Sheet by name or ID. It can also build an index for several fields to do quick rows lookup (see section "Custom Indexes")\n  * `def list()`: fetches a list of all sheets (summary only)\n  * `def create(obj: Sheet)`: adds a new sheet\n  * `def update(obj: Sheet)`: updates a sheet\n  * `def delete(name: Optional[str], id: Optional[int])`: deletes a sheet by name or ID\n\n#### Class `simple_smartsheet.models.Sheet`\nAttributes (converted from camelCase to snake_case):\n  * [http://smartsheet-platform.github.io/api-docs/#sheets](http://smartsheet-platform.github.io/api-docs/#sheets)\n  \nMethods:\n  * `def update_index()`: rebuilds mapping tables for rows and columns for quick lookup\n  * `def get_row(row_num: Optional[int], row_id: Optional[int], filter: Optional[Dict[str, Any]])`: returns a Row object by row number, ID or by filter, if a unique index was built (see section "Custom Indexes")\n  * `def get_rows(index_query: Dict[str, Any])`: returns list of Row objects by filter, if an index was built (see section "Custom Indexes")\n  * `def get_column(column_title: Optional[str], column_id: Optional[int])`: returns a Column object by column title or id\n  * `def add_rows(rows: Sequence[Row])`: adds rows to the sheet\n  * `def add_row(row: Row)`: add a single row to the sheet\n  * `def update_rows(rows: Sequence[Row])`: updates several rows in the sheet\n  * `def update_row(row: Row)`: updates a single row\n  * `def delete_rows(row_ids: Sequence[int])`: delete several rows with provided ids\n  * `def delete_row(row_id: int)`: delete a single row with a provided id\n  * `def sort_rows(order: List[Dict[str, Any]])`: sorts sheet rows with the specified order. An argument example:  \n```\n[\n    {"column_title": "Birth date", "descending": True},\n    {"column_title": "Full Name"}\n]\n```\n  * `def make_cell(column_title: str, field_value: Union[float, str, datetime, None])`: creates a Cell object with provided column title and an associated value\n  * `def make_cells(fields: Dict[str, Union[float, str, datetime, None]])`: creates a list of Cell objects from an input dictionary where column title is key associated with the field value\n  * `def as_list()`: returns a list of dictionaries where column title is key associated with the field value\n  \n#### Class `simple_smartsheet.models.Row`\nAttributes (converted from camelCase to snake_case):\n  * [http://smartsheet-platform.github.io/api-docs/#rows](http://smartsheet-platform.github.io/api-docs/#rows)\n  * `rowNumber` is mapped to `num`\n  \nMethods:\n  * `def get_cell(column_title: Optional[str], column_id: Optional[int])` - returns a Cell object by column title (case-sensitive) or column id\n  * `def as_dict()` - returns a dictionary of column title to cell value mappings\n\n#### Class `simple_smartsheet.models.Column`\nAttributes (converted from camelCase to snake_case):\n  * [http://smartsheet-platform.github.io/api-docs/#columns](http://smartsheet-platform.github.io/api-docs/#columns)\n\n#### Class `simple_smartsheet.models.Cell`\nAttributes (converted from camelCase to snake_case):\n  * [http://smartsheet-platform.github.io/api-docs/#cells](http://smartsheet-platform.github.io/api-docs/#cells)\n  \n### Custom Indexes\nWhen you are retrieving a smartsheet, it is possible to build an index to enable quick rows lookups.\nThis is controlled using `index_key` argument in `get` method. This argument is a dictionary with two keys `columns` and `unique`. `columns` should contain a tuple with column titles (case sensitive). `unique` controls if the index always points to a single row (value `True`, lookups are done using `get_row` method) or multiple rows (value `False`, lookups are done using `get_rows` method).\n\nBelow you can find a code example:\n```python\nfrom simple_smartsheet import Smartsheet\nfrom pprint import pprint\n\nTOKEN = \'my-token\'\nsmartsheet = Smartsheet(TOKEN)\n\nINDEX_KEYS = [\n    {"columns": ("Company Name",), "unique": False},\n    {"columns": ("Company Name", "Full Name"), "unique": True},\n    {"columns": ("Email Address",), "unique": True},\n]\nsheet = smartsheet.sheets.get("Index Test Sheet", index_keys=INDEX_KEYS)\n\npprint(sheet.indexes)\n# >\n# defaultdict(<class \'dict\'>,\n#             {(\'Company Name\',): {(\'ACME\',): [Row(id=525791232583556, num=1),\n#                                              Row(id=5029390859954052, num=2)],\n#                                  (\'Globex\',): [Row(id=2777591046268804, num=3)]},\n#              (\'Company Name\', \'Full Name\'): {(\'ACME\', \'Alice Smith\'): Row(id=525791232583556, num=1),\n#                                              (\'ACME\', \'Bob Lee\'): Row(id=5029390859954052, num=2),\n#                                              (\'Globex\', \'Charlie Brown\'): Row(id=2777591046268804, num=3)},\n#              (\'Email Address\',): {(\'alice.smith@acme.com\',): Row(id=525791232583556, num=1),\n#                                   (\'bob.lee@acme.com\',): Row(id=5029390859954052, num=2),\n#                                   (\'charlie.brown@globex.com\',): Row(id=2777591046268804, num=3)}})\n\npprint([row.as_dict() for row in sheet.rows])\n# >\n# [{\'Company Name\': \'ACME\',\n#   \'Email Address\': \'alice.smith@acme.com\',\n#   \'Full Name\': \'Alice Smith\'},\n#  {\'Company Name\': \'ACME\',\n#   \'Email Address\': \'bob.lee@acme.com\',\n#   \'Full Name\': \'Bob Lee\'},\n#  {\'Company Name\': \'Globex\',\n#   \'Email Address\': \'charlie.brown@globex.com\',\n#   \'Full Name\': \'Charlie Brown\'}]\n\npprint(sheet.get_row(filter={"Email Address": "charlie.brown@globex.com"}).as_dict())\n# >\n# {\'Company Name\': \'Globex\',\n#  \'Email Address\': \'charlie.brown@globex.com\',\n#  \'Full Name\': \'Charlie Brown\'}\n\npprint(\n    sheet.get_row(filter={"Full Name": "Alice Smith", "Company Name": "ACME"}).as_dict()\n)\n# >\n# {\'Company Name\': \'ACME\',\n#  \'Email Address\': \'alice.smith@acme.com\',\n#  \'Full Name\': \'Alice Smith\'}\n\npprint([row.as_dict() for row in sheet.get_rows(filter={"Company Name": "ACME"})])\n# >\n# [{\'Company Name\': \'ACME\',\n#   \'Email Address\': \'alice.smith@acme.com\',\n#   \'Full Name\': \'Alice Smith\'},\n#  {\'Company Name\': \'ACME\',\n#   \'Email Address\': \'bob.lee@acme.com\',\n#   \'Full Name\': \'Bob Lee\'}]\n``` ',
    'author': 'Dmitry Figol',
    'author_email': 'git@dmfigol.me',
    'url': 'https://github.com/dmfigol/simple-smartsheet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
